from datetime import datetime

import marshmallow
from marshmallow import validate, fields

from channelstream.validation import (
    BackportedDict,
    ChannelstreamSchema,
    gen_uuid,
    validate_connection_id,
    validate_username,
    UserStateField,
)


class ChannelConfigSchema(ChannelstreamSchema):
    notify_presence = fields.Boolean(
        missing=False, description="Should presence operations emit notifications"
    )
    store_history = fields.Boolean(
        missing=False, description="Should store history messages"
    )
    history_size = fields.Integer(
        missing=10,
        validate=[validate.Range(min=0)],
        description="How many messages should be stored",
    )
    broadcast_presence_with_user_lists = fields.Boolean(
        missing=False,
        description="should currently connected users be sent on connection requests",
    )
    notify_state = fields.Boolean(
        missing=False, description="Should emit public state changes"
    )
    store_frames = fields.Boolean(
        missing=True, description="Should store catchup frames"
    )


class InfoResolutionSchema(ChannelstreamSchema):
    include_history = fields.Boolean(missing=True)
    include_users = fields.Boolean(missing=True)
    channels = fields.List(fields.String(validate=validate.Length(min=1, max=256)))
    exclude_channels = fields.List(
        fields.String(validate=validate.Length(min=1, max=256))
    )
    include_connections = fields.Boolean(missing=False)
    return_public_state = fields.Boolean(missing=False)


class ChannelInfoBodySchema(ChannelstreamSchema):
    info = fields.Nested(InfoResolutionSchema, required=True)


class ConnectBodySchema(ChannelstreamSchema):
    username = fields.String(
        required=True,
        validate=validate.Length(min=1, max=512),
        description="Identifier for user, will be used for other operations",
    )

    conn_id = fields.UUID(
        missing=gen_uuid,
        description="Connection identifier used for broadcast listening",
    )

    channels = fields.List(
        fields.String(validate=validate.Length(min=1, max=256)),
        description="List of channels user should be subscribed to, "
        "the chnnels will be auto created upon connection",
        missing=lambda: [],
    )

    state_public_keys = fields.List(
        fields.String(description="", validate=validate.Length(min=1, max=256)),
        missing=lambda: [],
        description="What state keys should be visible/emitted to other users",
    )

    fresh_user_state = BackportedDict(
        missing=lambda: {},
        values=UserStateField(allow_none=True),
        keys=fields.String(),
        description="Default user state if user object is not in server memory",
    )
    user_state = BackportedDict(
        missing=lambda: {},
        values=UserStateField(allow_none=True),
        keys=fields.String(),
        description="Update user state to values in this dictionary, "
        "values not present in dict are kept intact, valid values are: "
        "string, int, float and boolean",
    )
    channel_configs = BackportedDict(
        missing=lambda: {},
        values=fields.Nested(ChannelConfigSchema()),
        keys=fields.String(),
        description="Sets configuration for newly created channels "
        "in form of channelName:ChannelConfigBody",
    )
    info = fields.Nested(
        InfoResolutionSchema(),
        missing=lambda: {},
        description="Controls how much information should be returned in response",
    )


class SubscribeBodySchema(ChannelstreamSchema):
    conn_id = fields.UUID(required=True, validate=[validate_connection_id])

    channels = fields.List(
        fields.String(validate=validate.Length(min=1, max=256)),
        required=True,
        validate=validate.Length(min=1),
        description="List of channels user should be subscribed to",
    )
    info = fields.Nested(InfoResolutionSchema(), missing=lambda: {})

    channel_configs = BackportedDict(
        missing=lambda: {},
        values=fields.Nested(ChannelConfigSchema()),
        keys=fields.String(),
        description="Sets configuration for newly created channels "
        "in form of channelName:ChannelConfigBody",
    )

    @marshmallow.pre_load
    def get_connection(self, in_data):
        in_data.setdefault("conn_id", self.context["request"].GET.get("conn_id"))
        return in_data


class UnsubscribeBodySchema(SubscribeBodySchema):
    pass


class UserStateBodySchema(ChannelstreamSchema):
    user = fields.String(
        required=True, validate=[validate.Length(min=1, max=512), validate_username]
    )

    user_state = BackportedDict(
        missing=lambda: {},
        values=UserStateField(allow_none=True),
        keys=fields.String(),
        description="Update user state to values in this dictionary, "
        "values not present in dict are kept intact, valid values are: "
        "string, int, float and boolean",
    )
    state_public_keys = fields.List(
        fields.String(validate=validate.Length(min=1, max=256)),
        missing=None,
        description="What state keys should be visible/emitted to other users",
    )


class PayloadDeliveryInfo(ChannelstreamSchema):
    pm_users = fields.List(
        fields.String(validate=validate.Length(min=1, max=512)),
        missing=lambda: [],
        description="Which users should get the message, either globally or in channel context",
    )
    exclude_users = fields.List(
        fields.String(validate=validate.Length(min=1, max=512)),
        missing=lambda: [],
        description="Which users should not have the message delivered, preceeds pm_users",
    )
    channel = fields.String(
        validate=validate.Length(min=1, max=256),
        missing=None,
        description="What channel should the message be delivered to",
    )
    no_history = fields.Boolean(
        missing=False, description="Should store the message in history"
    )


class MessageBodySchema(PayloadDeliveryInfo, ChannelstreamSchema):
    uuid = fields.UUID(
        default=gen_uuid, missing=gen_uuid, description="Identifier of emitted message"
    )
    timestamp = fields.DateTime(
        missing=lambda: datetime.utcnow().isoformat(),
        description="Registered timestamp of message",
    )
    user = fields.String(
        required=True,
        validate=validate.Length(min=1, max=512),
        description="Sender identifier",
    )
    message = fields.Dict(description="Message payload, can hold other keys")

    @marshmallow.post_load()
    def _add_unknown(self, data):
        data["edited"] = None
        return data


class MessageEditBodySchema(MessageBodySchema):
    uuid = fields.UUID(
        required=True, description="Identifier of message that is being edited"
    )
    timestamp = fields.DateTime()
    user = fields.String(validate=validate.Length(min=1, max=512))
    message = fields.Dict()
    edited = fields.DateTime(missing=lambda: datetime.utcnow().isoformat())

    @marshmallow.post_load()
    def _add_unknown(self, data):
        return data


class MessagesDeleteBodySchema(PayloadDeliveryInfo, ChannelstreamSchema):
    uuid = fields.UUID(
        required=True,
        description="Identifier of message that should be deleted from history",
    )


class DisconnectBodySchema(ChannelstreamSchema):
    conn_id = fields.UUID(required=True)



SHARED_DEFAULTS = {
    "secret": "secret",
    "admin_user": "admin",
    "admin_secret": "admin_secret",
    "cookie_secret": "",
    "gc_conns_after": 30,
    "gc_channels_after": 3600 * 72,
    "wake_connections_after": 5,
    "allow_posting_from": ["127.0.0.1"],
    "port": 8000,
    "host": "0.0.0.0",
    "debug": False,
    "log_level": "INFO",
    "demo": False,
    "allow_cors": "",
    "validate_requests": True,
    "enforce_https": False,
    "http_scheme": "",
}

CONFIGURABLE_PARAMS = (
    "debug",
    "log_level",
    "port",
    "host",
    "secret",
    "admin_user",
    "admin_secret",
    "cookie_secret",
    "allow_posting_from",
    "allow_cors",
    "validate_requests",
    "enforce_https",
    "http_scheme",
)



import logging
from datetime import datetime, timedelta

import gevent

from channelstream.server_state import get_state

log = logging.getLogger(__name__)


def gc_conns():
    server_state = get_state()
    with server_state.lock:
        start_time = datetime.utcnow()
        threshold = start_time - timedelta(seconds=15)
        collected_conns = []
        # collect every ref in chanels
        # remove connections from channels
        for channel in server_state.channels.values():
            for username, conns in list(channel.connections.items()):
                for conn in conns:
                    if conn.last_active < threshold:
                        channel.connections[username].remove(conn)
                        collected_conns.append(conn)
                channel.after_parted(username)
        # remove old conns from users and connection dictionaries
        for conn in collected_conns:
            if conn.username in server_state.users:
                if conn in server_state.users[conn.username].connections:
                    server_state.users[conn.username].connections.remove(conn)
            if conn.id in server_state.connections:
                del server_state.connections[conn.id]
            # make sure connection is closed after we garbage
            # collected it from our list
            conn.queue = None
            if conn.socket:
                try:
                    conn.socket.close()
                except Exception:
                    raise
        log.debug(
            "gc_conns() removed:%s time %s"
            % (len(collected_conns), datetime.utcnow() - start_time)
        )


def gc_users():
    server_state = get_state()
    with server_state.lock:
        counter = 0
        start_time = datetime.utcnow()
        threshold = datetime.utcnow() - timedelta(days=1)
        for user in list(server_state.users.values()):
            if user.last_active < threshold:
                counter += 1
                server_state.users.pop(user.username)
        log.debug(
            "gc_users() removed:%s time %s" % (counter, datetime.utcnow() - start_time)
        )


def gc_users_forever():
    while True:
        try:
            gc_users()
        except Exception as exc:
            log.error(exc)
        gevent.sleep(60)


def gc_conns_forever():
    while True:
        try:
            gc_conns()
        except Exception as exc:
            log.error(exc)
        gevent.sleep(1)



import argparse
import copy
import logging
import json
import pkg_resources
import jinja2
import os

from channelstream.cli import CONFIGURABLE_PARAMS, SHARED_DEFAULTS

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def main():
    config = copy.deepcopy(SHARED_DEFAULTS)

    parser = argparse.ArgumentParser(add_help=True)
    parser.add_argument(
        "operation", help="Operation", default=None, choices=["make_config"]
    )
    parser.add_argument("-j", "--json", dest="json", help="Config JSON", default=None)
    parser.add_argument(
        "-o", "--output", dest="output", help="Output file", required=True
    )
    args = parser.parse_args()
    data_json = {}
    if args.json:
        data_json = json.loads(args.json)

    for key in CONFIGURABLE_PARAMS:
        conf_value = data_json.get(key)
        if conf_value:
            config[key] = conf_value

    if args.operation == "make_config":
        template_path = os.path.join("templates", "ini", "channelstream.ini.jinja2")
        template_str = pkg_resources.resource_string("channelstream", template_path)
        template = jinja2.Template(template_str.decode("utf8"))
        template_vars = config
        compiled = template.render(**template_vars)
        with open(args.output, "w") as f:
            f.write(compiled)
        log.info("Config written")



import logging
from datetime import datetime

import gevent
import gevent.util
from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from gevent.queue import Queue, Empty
from pyramid.httpexceptions import HTTPUnauthorized, HTTPFound
from pyramid.security import remember, forget, NO_PERMISSION_REQUIRED
from pyramid.view import view_config, view_defaults
from pyramid_apispec.helpers import add_pyramid_paths

from channelstream import operations, utils, patched_json as json, __version__
from channelstream.server_state import get_state, STATS
from channelstream.validation import schemas

log = logging.getLogger(__name__)


class SharedUtils(object):
    def __init__(self, request):
        self.request = request

    def get_channel_info(
        self,
        req_channels=None,
        include_history=True,
        include_connections=False,
        include_users=False,
        exclude_channels=None,
        return_public_state=False,
    ):
        """
        Gets channel information for req_channels or all channels
        if req_channels is None
        :param: include_history (bool) will include message history
                for the channel
        :param: include_connections (bool) will include connection list
                for users
        :param: include_users (bool) will include user list for the channel
        :param: exclude_channels (bool) will exclude specific channels
                from info list (handy to exclude global broadcast)
        """
        server_state = get_state()
        if not exclude_channels:
            exclude_channels = []
        start_time = datetime.utcnow()

        json_data = {"channels": {}, "users": []}

        users_to_list = set()

        # select everything for empty list
        if req_channels is None:
            channel_instances = server_state.channels.values()
        else:
            channel_instances = [
                server_state.channels[c]
                for c in req_channels
                if c in server_state.channels
            ]

        for channel_inst in channel_instances:
            if channel_inst.name in exclude_channels:
                continue

            channel_info = channel_inst.get_info(
                include_history=include_history, include_users=include_users
            )
            json_data["channels"][channel_inst.name] = channel_info
            users_to_list.update(channel_info["users"])

        for username in users_to_list:
            user = server_state.users[username]
            json_data["users"].append(
                {
                    "user": username,
                    "state": user.state
                    if not return_public_state
                    else user.public_state,
                }
            )
        log.info("info time: %s" % (datetime.utcnow() - start_time))
        return json_data

    def get_common_info(self, channels, info_config):
        """
        Return channel information based on requirements
        :param channels:
        :param info_config:
        :return:
        """
        include_history = info_config.get("include_history", True)
        include_users = info_config.get("include_users", True)
        exclude_channels = info_config.get("exclude_channels", [])
        include_connections = info_config.get("include_connections", False)
        return_public_state = info_config.get("return_public_state", False)
        channels_info = self.get_channel_info(
            channels,
            include_history=include_history,
            include_connections=include_connections,
            include_users=include_users,
            exclude_channels=exclude_channels,
            return_public_state=return_public_state,
        )
        return channels_info


@view_config(route_name="connect", request_method="POST", renderer="json")
def connect(request):
    """
    Connect view
    ---
    post:
      security:
        - APIKeyHeader: []
      tags:
      - "API"
      summary: "connects users to the server"
      description: ""
      operationId: "connect"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        required: true
        schema:
          $ref: "#/definitions/ConnectBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
          schema:
            $ref: '#/definitions/ConnectBody'
    """
    shared_utils = SharedUtils(request)
    schema = schemas.ConnectBodySchema(context={"request": request})
    json_body = schema.load(request.json_body).data
    channels = sorted(json_body["channels"])
    connection, user = operations.connect(
        username=json_body["username"],
        fresh_user_state=json_body["fresh_user_state"],
        state_public_keys=json_body["state_public_keys"],
        update_user_state=json_body["user_state"],
        conn_id=json_body["conn_id"],
        channels=channels,
        channel_configs=json_body["channel_configs"],
    )

    # get info config for channel information
    channels_info = shared_utils.get_common_info(channels, json_body["info"])

    return {
        "conn_id": connection.id,
        "state": user.state,
        "username": user.username,
        "public_state": user.public_state,
        "channels": channels,
        "channels_info": channels_info,
    }


@view_config(route_name="subscribe", request_method="POST", renderer="json")
def subscribe(request):
    """
    Subscribe view
    ---
    post:
      security:
        - APIKeyHeader: []
      tags:
      - "API"
      summary: "Subscribes connection to new channels"
      description: ""
      operationId: "subscribe"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        required: true
        schema:
          $ref: "#/definitions/SubscribeBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """
    server_state = get_state()
    shared_utils = SharedUtils(request)
    schema = schemas.SubscribeBodySchema(context={"request": request})
    json_body = schema.load(request.json_body).data
    connection = server_state.connections.get(json_body["conn_id"])
    channels = json_body["channels"]
    channel_configs = json_body.get("channel_configs", {})
    subscribed_to = operations.subscribe(
        connection=connection, channels=channels, channel_configs=channel_configs
    )

    # get info config for channel information
    current_channels = connection.channels
    channels_info = shared_utils.get_common_info(current_channels, json_body["info"])
    return {
        "channels": current_channels,
        "channels_info": channels_info,
        "subscribed_to": sorted(subscribed_to),
    }


@view_config(route_name="unsubscribe", request_method="POST", renderer="json")
def unsubscribe(request):
    """
    Unsubscribe view
    ---
    post:
      security:
        - APIKeyHeader: []
      tags:
      - "API"
      summary: "Removes connection from channels"
      description: ""
      operationId: "unsubscribe"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        required: true
        schema:
          $ref: "#/definitions/UnsubscribeBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """
    server_state = get_state()
    shared_utils = SharedUtils(request)
    schema = schemas.UnsubscribeBodySchema(context={"request": request})
    json_body = schema.load(request.json_body).data
    connection = server_state.connections.get(json_body["conn_id"])
    unsubscribed_from = operations.unsubscribe(
        connection=connection, unsubscribe_channels=json_body["channels"]
    )

    # get info config for channel information
    current_channels = connection.channels
    channels_info = shared_utils.get_common_info(current_channels, json_body["info"])
    return {
        "channels": current_channels,
        "channels_info": channels_info,
        "unsubscribed_from": sorted(unsubscribed_from),
    }


@view_config(
    route_name="api_listen",
    request_method="GET",
    renderer="json",
    permission=NO_PERMISSION_REQUIRED,
)
def listen(request):
    """
    Handles long polling connections
    ---
    get:
      tags:
      - "Client API"
      summary: "Handles long polling connections"
      description: ""
      operationId: "listen"
      produces:
      - "application/json"
      responses:
        200:
          description: "Success"
    """
    server_state = get_state()
    config = request.registry.settings
    conn_id = utils.uuid_from_string(request.params.get("conn_id"))
    connection = server_state.connections.get(conn_id)
    if not connection:
        raise HTTPUnauthorized()
    # attach a queue to connection
    connection.queue = Queue()
    connection.deliver_catchup_messages()
    request.response.app_iter = yield_response(request, connection, config)
    return request.response


def yield_response(request, connection, config):
    messages = await_data(connection, config)
    connection.mark_activity()
    cb = request.params.get("callback")
    if cb:
        resp = cb + "(" + json.dumps(messages) + ")"
    else:
        resp = json.dumps(messages)
    yield resp.encode("utf8")


def await_data(connection, config):
    messages = []
    # block for first message - wake up after a while
    try:
        messages.extend(connection.queue.get(timeout=config["wake_connections_after"]))
    except Empty:
        pass
    # get more messages if enqueued takes up total 0.25
    while True:
        try:
            messages.extend(connection.queue.get(timeout=0.25))
        except Empty:
            break
    return messages


@view_config(route_name="user_state", request_method="POST", renderer="json")
def user_state(request):
    """
    Sets the state of a user object
    ---
    post:
      security:
        - APIKeyHeader: []
      tags:
      - "API"
      summary: "set the status of specific user"
      description: ""
      operationId: "user_state"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        required: true
        schema:
          $ref: "#/definitions/UserStateBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """
    server_state = get_state()
    schema = schemas.UserStateBodySchema(context={"request": request})
    data = schema.load(request.json_body).data
    user_inst = server_state.users[data["user"]]
    # can be empty list!
    if data["state_public_keys"] is not None:
        user_inst.state_public_keys = data["state_public_keys"]
    changed = operations.change_user_state(
        user_inst=user_inst, user_state=data["user_state"]
    )
    return {
        "user_state": user_inst.state,
        "changed_state": changed,
        "public_keys": user_inst.state_public_keys,
    }


def shared_messages(request):
    server_state = get_state()
    schema = schemas.MessageBodySchema(context={"request": request}, many=True)
    data = schema.load(request.json_body).data
    data = [m for m in data if m.get("channel") or m.get("pm_users")]
    for msg in data:
        gevent.spawn(operations.pass_message, msg, server_state.stats)
    return list(data)


# prepare v1 version
# @view_config(route_name="api_v1_messages", request_method="POST", renderer="json")
def messages_post(request):
    """
    Send message to channels and/or users
    ---
    post:
      security:
        - APIKeyHeader: []
      tags:
      - "V1 API (future stable)"
      summary: "Send message to channels and/or users"
      description: ""
      operationId: "message"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        required: true
        schema:
          $ref: "#/definitions/MessagesBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """
    return shared_messages(request)


@view_config(route_name="message", request_method="POST", renderer="json")
def message(request):
    """
    Send message to channels and/or users
    ---
    post:
      security:
        - APIKeyHeader: []
      tags:
      - "API"
      summary: "Send message to channels and/or users"
      description: ""
      operationId: "message"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        required: true
        schema:
          $ref: "#/definitions/MessagesBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """
    return shared_messages(request)


@view_config(route_name="message", request_method="PATCH", renderer="json")
def messages_patch(request):
    """
    Edit existing message in history and emit changes
    ---
    patch:
      security:
        - APIKeyHeader: []
      tags:
      - "API"
      summary: "Edit existing message in history and emit changes"
      description: ""
      operationId: "edit_messages"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        required: true
        schema:
          $ref: "#/definitions/MessageEditBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """

    schema = schemas.MessageEditBodySchema(context={"request": request}, many=True)
    data = schema.load(request.json_body).data
    for msg in data:
        gevent.spawn(operations.edit_message, msg)
    return data


@view_config(route_name="message", request_method="DELETE", renderer="json")
def messages_delete(request):
    """
    Delete message from history and emit changes
    ---
    delete:
      security:
        - APIKeyHeader: []
      tags:
      - "API"
      summary: "Delete message from history and  emit changes"
      description: ""
      operationId: "messages_delete"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        required: true
        schema:
          $ref: "#/definitions/MessagesDeleteBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """

    schema = schemas.MessagesDeleteBodySchema(context={"request": request}, many=True)
    data = schema.load(request.json_body).data
    for msg in data:
        gevent.spawn(operations.delete_message, msg)
    return data


@view_config(
    route_name="api_disconnect", renderer="json", permission=NO_PERMISSION_REQUIRED
)
def disconnect(request):
    """
    Permanently remove connection from server
    ---
    get:
      tags:
      - "Client API"
      summary: "Permanently remove connection from server"
      description: ""
      operationId: "disconnect"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: query
        schema:
          type: string
        name: "conn_id"
        description: "Connection Id"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    post:
      tags:
      - "Client API"
      summary: "Permanently remove connection from server"
      description: ""
      operationId: "disconnect"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        schema:
          $ref: "#/definitions/DisconnectBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """
    schema = schemas.DisconnectBodySchema(context={"request": request})
    if request.method != "POST":
        payload = {"conn_id": request.GET.get("conn_id")}
    else:
        json_body = request.json_body
        payload = {"conn_id": json_body.get("conn_id")}
    data = schema.load(payload).data
    return operations.disconnect(conn_id=data["conn_id"])


@view_config(route_name="channel_config", request_method="POST", renderer="json")
def channel_config(request):
    """
    Set channel configuration
    ---
    post:
      security:
        - APIKeyHeader: []
      tags:
      - "API"
      summary: "Set channel configuration"
      description: ""
      operationId: "channel_config"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        schema:
          $ref: "#/definitions/ChannelConfigBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """

    shared_utils = SharedUtils(request)

    deserialized = {}
    schema = schemas.ChannelConfigSchema(context={"request": request})
    json_body = request.json_body
    for k in json_body.keys():
        deserialized[k] = schema.load(json_body[k]).data
    operations.set_channel_config(channel_configs=deserialized)
    channels_info = shared_utils.get_channel_info(
        deserialized.keys(), include_history=False, include_users=False
    )
    return channels_info


@view_config(route_name="info", renderer="json")
def info(request):
    """
    Returns channel information
    ---
    post:
      security:
        - APIKeyHeader: []
      tags:
      - "API"
      summary: "Returns channel information"
      description: ""
      operationId: "info"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Request JSON body"
        schema:
          $ref: "#/definitions/ChannelInfoBody"
      responses:
        422:
          description: "Unprocessable Entity"
        200:
          description: "Success"
    """
    server_state = get_state()
    shared_utils = SharedUtils(request)
    if not request.body:
        req_channels = server_state.channels.keys()
        info_config = {
            "include_history": True,
            "include_users": True,
            "exclude_channels": [],
            "include_connections": True,
        }
    else:
        schema = schemas.ChannelInfoBodySchema(context={"request": request})
        data = schema.load(request.json_body).data
        # get info config for channel information
        info_config = data.get("info") or {}
        req_channels = info_config.get("channels", None)
        info_config["include_connections"] = info_config.get(
            "include_connections", True
        )
    channels_info = shared_utils.get_common_info(req_channels, info_config)
    return channels_info


@view_defaults(route_name="action", renderer="json", permission="admin")
class ServerViews(object):
    def __init__(self, request):
        self.request = request
        self.utils = SharedUtils(request)

    @view_config(route_name="admin", renderer="templates/admin.jinja2")
    def admin(self):
        """
        Serve admin page html
        :return:
        """
        return {}

    @view_config(
        route_name="admin_action", match_param=("action=debug",), renderer="string"
    )
    def admin_debug(self):
        return "\n".join(gevent.util.format_run_info())

    @view_config(
        route_name="admin_action",
        match_param=("action=sign_in",),
        renderer="templates/sign_in.jinja2",
        permission=NO_PERMISSION_REQUIRED,
    )
    def admin_sign_in(self):
        if self.request.method == "POST":
            admin_user = self.request.registry.settings["admin_user"]
            admin_secret = self.request.registry.settings["admin_secret"]
            username = self.request.POST.get("username", "").strip()
            password = self.request.POST.get("password", "").strip()
            if username == admin_user and password == admin_secret:
                headers = remember(self.request, admin_user)
                url = self.request.route_url("admin")
                return HTTPFound(url, headers=headers)
            else:
                # make potential brute forcing non-feasible
                gevent.sleep(0.5)
        return {}

    @view_config(
        route_name="admin_action",
        match_param=("action=sign_out",),
        renderer="string",
        permission=NO_PERMISSION_REQUIRED,
    )
    def admin_sign_out(self):
        headers = forget(self.request)
        url = self.request.route_url("admin_action", action="sign_in")
        return HTTPFound(url, headers=headers)

    @view_config(
        route_name="admin_json", renderer="json", request_method=("POST", "GET")
    )
    def admin_json(self):
        """
        Admin json
        ---
        get:
          tags:
          - "Admin API"
          summary: "Return server information in json format for admin panel
          purposes"
          description: ""
          operationId: "admin_json"
          consumes:
          - "application/json"
          produces:
          - "application/json"
          parameters:
          - in: "body"
            name: "body"
            description: "Response info configuration"
          responses:
            422:
              description: "Unprocessable Entity"
            200:
              description: "Success"
        post:
          tags:
          - "Admin API"
          summary: "Return server information in json format for admin panel
          purposes"
          description: ""
          operationId: "admin_json"
          consumes:
          - "application/json"
          produces:
          - "application/json"
          parameters:
          - in: "body"
            name: "body"
            description: "Response info configuration"
          responses:
            422:
              description: "Unprocessable Entity"
            200:
              description: "Success"
        """
        server_state = get_state()
        uptime = datetime.utcnow() - STATS["started_on"]
        uptime = str(uptime).split(".")[0]
        remembered_user_count = len([user for user in server_state.users.items()])
        active_users = [
            user for user in server_state.users.values() if user.connections
        ]
        unique_user_count = len(active_users)
        total_connections = sum([len(user.connections) for user in active_users])
        channels_info = self.utils.get_common_info(
            None,
            {
                "include_history": True,
                "include_users": True,
                "exclude_channels": [],
                "include_connections": True,
            },
        )
        return {
            "remembered_user_count": remembered_user_count,
            "unique_user_count": unique_user_count,
            "total_connections": total_connections,
            "total_channels": len(server_state.channels.keys()),
            "total_messages": server_state.stats["total_messages"],
            "total_unique_messages": server_state.stats["total_unique_messages"],
            "channels": channels_info["channels"],
            "users": [user.get_info(include_connections=True) for user in active_users],
            "uptime": uptime,
            "version": str(__version__),
        }

    @view_config(route_name="openapi_spec", renderer="json")
    def api_spec(self):
        """
        OpenApi 2.0 spec
        ---
        get:
          tags:
          - "OpenApi 2.0 spec"
          summary: "Return openapi spec
          purposes"
          description: ""
          operationId: "api_spec"
          consumes:
          - "application/json"
          produces:
          - "application/json"
          parameters:
          responses:
            200:
              description: "Success"
        """
        spec = APISpec(
            title="Channelstream API",
            version="0.7.0",
            openapi_version="2.0.0",
            plugins=(MarshmallowPlugin(),),
        )
        spec.components.schema("ConnectBody", schema=schemas.ConnectBodySchema)
        spec.components.schema("SubscribeBody", schema=schemas.SubscribeBodySchema)
        spec.components.schema("UnsubscribeBody", schema=schemas.UnsubscribeBodySchema)
        spec.components.schema("UserStateBody", schema=schemas.UserStateBodySchema)
        spec.components.schema(
            "MessagesBody", schema=schemas.MessageBodySchema(many=True)
        )
        spec.components.schema("MessageBody", schema=schemas.MessageBodySchema())
        spec.components.schema(
            "MessageEditBody", schema=schemas.MessageEditBodySchema(many=True)
        )
        spec.components.schema(
            "MessagesDeleteBody", schema=schemas.MessagesDeleteBodySchema(many=True)
        )
        spec.components.schema("DisconnectBody", schema=schemas.DisconnectBodySchema)
        spec.components.schema("ChannelConfigBody", schema=schemas.ChannelConfigSchema)
        spec.components.schema("ChannelInfoBody", schema=schemas.ChannelInfoBodySchema)

        # api
        add_pyramid_paths(spec, "connect", request=self.request)
        add_pyramid_paths(spec, "subscribe", request=self.request)
        add_pyramid_paths(spec, "unsubscribe", request=self.request)
        add_pyramid_paths(spec, "user_state", request=self.request)
        add_pyramid_paths(spec, "message", request=self.request)
        add_pyramid_paths(spec, "channel_config", request=self.request)
        add_pyramid_paths(spec, "info", request=self.request)

        add_pyramid_paths(spec, "api_listen", request=self.request)
        add_pyramid_paths(spec, "api_listen_ws", request=self.request)
        add_pyramid_paths(spec, "api_disconnect", request=self.request)

        add_pyramid_paths(spec, "admin_json", request=self.request)
        spec_dict = spec.to_dict()
        spec_dict["securityDefinitions"] = {
            "APIKeyHeader": {
                "type": "apiKey",
                "name": "X-Channelstream-Secret",
                "in": "header",
            }
        }
        return spec_dict



import copy
import logging
import uuid
from datetime import datetime

from channelstream.server_state import get_state
from channelstream.utils import process_catchup
from channelstream.validation import MSG_EDITABLE_KEYS

log = logging.getLogger(__name__)


class Channel(object):
    """ Represents one of our chat channels - has some config options """

    config_keys = [
        "notify_presence",
        "store_history",
        "history_size",
        "broadcast_presence_with_user_lists",
        "notify_state",
        "store_frames",
    ]

    def __init__(self, name, long_name=None, channel_config=None):
        """

        :param name:
        :param long_name:
        :param channel_config:
        """
        self.uuid = uuid.uuid4()
        self.name = name
        self.long_name = long_name
        self.last_active = None
        self.connections = {}
        self.notify_presence = False
        self.broadcast_presence_with_user_lists = False
        # channel sends all user state key changes
        self.notify_state = False
        self.salvageable = False
        self.store_history = False
        self.store_frames = True
        self.history_size = 10
        self.history = []
        # store frames for fetching when connection is established
        # those frames will store channel messages including presence ones
        self.frames = []
        if channel_config:
            self.reconfigure_from_dict(channel_config)
        log.info("%s created" % self)
        log.info("Configuration used: {}".format(channel_config))
        self.mark_activity()

    def mark_activity(self):
        self.last_active = datetime.utcnow()

    def get_catchup_frames(self, newer_than, username):
        found = []
        for t, f in self.frames:
            # either old frame or user is excluded or PM not meant for user
            if (
                t < newer_than
                or (f["exclude_users"] and username in f["exclude_users"])
                or (f["pm_users"] and username not in f["pm_users"])
            ):
                continue

            found.append(process_catchup(f))
        return found

    def reconfigure_from_dict(self, config):
        if config:
            for key in self.config_keys:
                val = config.get(key)
                if val is not None:
                    setattr(self, key, val)

    def add_connection(self, connection):
        username = connection.username
        connections = self.connections.setdefault(username, [])
        if not connections and self.notify_presence:
            self.send_notify_presence_info(username, "joined")
        if connection not in connections:
            connections.append(connection)
            return True
        return False

    def remove_connection(self, connection):
        was_found = False
        username = connection.username
        connections = self.connections.setdefault(username, [])
        if connection in connections:
            self.connections[username].remove(connection)
            was_found = True

        self.after_parted(username)
        return was_found

    def after_parted(self, username):
        """
        Sends parted message if necessary and removed username from
        connections if empty

        :param username:
        :return:
        """
        if not self.connections[username]:
            del self.connections[username]
            if self.notify_presence:
                self.send_notify_presence_info(username, "parted")

    def send_notify_presence_info(self, username, action):
        """
        Sends a message to other connected parties about a presence change
        :param username:
        :param action:
        :return:
        """
        server_state = get_state()
        connected_users = []
        if self.broadcast_presence_with_user_lists:
            for _username in self.connections.keys():
                user_inst = server_state.users.get(_username)
                user_data = {
                    "user": user_inst.username,
                    "state": user_inst.public_state,
                }
                connected_users.append(user_data)

        self.mark_activity()
        payload = {
            "uuid": uuid.uuid4(),
            "type": "presence",
            "no_history": False,
            "pm_users": [],
            "exclude_users": [username],
            "user": username,
            "users": connected_users,
            "timestamp": self.last_active,
            "channel": self.name,
            "message": {"action": action},
            "state": None,
            "catchup": False,
        }
        if action == "joined":
            payload["state"] = server_state.users[username].public_state
        self.add_message(payload, exclude_users=payload["exclude_users"])
        return payload

    def send_user_state(self, user_inst, changed):
        self.mark_activity()

        public_changed = [x for x in changed if x["key"] in user_inst.public_state]

        payload = {
            "uuid": uuid.uuid4(),
            "type": "user_state_change",
            "no_history": False,
            "pm_users": [],
            "exclude_users": [],
            "user": user_inst.username,
            "timestamp": self.last_active,
            "catchup": False,
            "channel": self.name,
            "message": {"state": user_inst.public_state, "changed": public_changed},
        }
        self.add_message(payload)
        return payload

    def add_frame(self, frame):
        if self.store_frames:
            self.frames.append((datetime.utcnow(), frame))
            self.frames = self.frames[-100:]

    def add_to_history(self, message):
        if self.store_history and message["type"] == "message":
            self.history.append(message)
            self.history = self.history[self.history_size * -1 :]

    def add_message(self, message, pm_users=None, exclude_users=None):
        """
        Sends the message to all connections subscribed to this channel
        """
        pm_users = pm_users or []
        exclude_users = exclude_users or []
        self.mark_activity()
        if not message["no_history"]:
            self.add_to_history(message)
        self.add_frame(message)
        message = copy.deepcopy(message)
        # do not leak delivery info
        del message["no_history"]
        del message["pm_users"]
        del message["exclude_users"]
        total_sent = 0
        # message everyone subscribed except excluded
        for user, conns in self.connections.items():
            if not exclude_users or user not in exclude_users:
                for connection in conns:
                    if not pm_users or connection.username in pm_users:
                        connection.add_message(message)
                        total_sent += 1
        return total_sent

    def __repr__(self):
        return "<Channel: %s, connections:%s>" % (self.name, len(self.connections))

    def get_info(self, include_history=True, include_users=False):
        server_state = get_state()
        settings = {k: getattr(self, k) for k in self.config_keys}

        chan_info = {
            "uuid": self.uuid,
            "name": self.name,
            "long_name": self.long_name,
            "settings": settings,
            "history": self.history if include_history else [],
            "last_active": self.last_active,
            "total_connections": sum(
                [len(conns) for conns in self.connections.values()]
            ),
            "total_users": 0,
            "users": [],
        }

        for username in self.connections.keys():
            user_inst = server_state.users.get(username)
            if include_users and user_inst.username not in chan_info["users"]:
                chan_info["users"].append(user_inst.username)
        chan_info["users"] = sorted(chan_info["users"])
        chan_info["total_users"] = len(chan_info["users"])
        return chan_info

    def alter_message(self, to_edit):
        found = False
        for msg in self.history:
            if msg["uuid"] == to_edit["uuid"]:
                found = True
                msg.update({k: v for k, v in to_edit.items() if k in MSG_EDITABLE_KEYS})
                break
        # if found history then reference in frames will be also updated,
        # otherwise search frames for channels that do not store history
        if not found:
            for f, msg in self.frames:
                if msg["uuid"] == to_edit["uuid"] and msg["type"] == "message":
                    msg.update(
                        {k: v for k, v in to_edit.items() if k in MSG_EDITABLE_KEYS}
                    )
                    break
        altered = copy.deepcopy(to_edit)
        altered["type"] = "message:edit"
        self.add_message(
            altered,
            pm_users=altered["pm_users"],
            exclude_users=altered["exclude_users"],
        )

    def delete_message(self, to_delete):
        for i, msg in enumerate(self.history):
            if msg["uuid"] == to_delete["uuid"]:
                self.history.pop(i)
                break

        for i, frame in enumerate(self.frames):
            msg = frame[1]
            if msg["uuid"] == to_delete["uuid"] and msg["type"] == "message":
                self.frames.pop(i)
                break

        deleted = copy.deepcopy(to_delete)
        deleted["type"] = "message:delete"
        self.add_message(
            deleted,
            pm_users=deleted["pm_users"],
            exclude_users=deleted["exclude_users"],
        )

    def __json__(self, request=None):
        return self.get_info()



from gevent import monkey

monkey.patch_all()

import copy
import pytest

from webtest import TestApp
from channelstream.cli import SHARED_DEFAULTS
from channelstream.cli.start import RoutingApplication


@pytest.mark.usefixtures("cleanup_globals")
class TestServerConfigView(object):
    def test_defaults(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        application = RoutingApplication(config)
        app = TestApp(application, extra_environ={"REMOTE_ADDR": "127.0.0.1"})
        app.get("/", status=404)
        app.post("/connect", status=403)

    def test_enforce_https_wrong(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        config["enforce_https"] = True
        application = RoutingApplication(config)
        app = TestApp(application, extra_environ={"REMOTE_ADDR": "127.0.0.1"})
        app.get("/", status=406)
        app.post("/connect", status=406)

    def test_enforce_https_proper(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        config["enforce_https"] = True
        application = RoutingApplication(config)
        app = TestApp(
            application,
            extra_environ={"REMOTE_ADDR": "127.0.0.1", "wsgi.url_scheme": "https"},
        )
        app.get("/", status=404)
        app.post("/connect", status=403)



"""View handlers package.
"""
import logging

from pyramid.security import NO_PERMISSION_REQUIRED

log = logging.getLogger(__name__)


def add_v1_routes(config):
    config.add_route("api_v1_messages", "/messages")


def includeme(config):
    config.add_static_view("static", path="channelstream:static/")
    config.add_route("CORS_route", "/*foo", request_method="OPTIONS")
    config.add_view(
        handle_CORS,
        route_name="CORS_route",
        renderer="string",
        permission=NO_PERMISSION_REQUIRED,
    )
    config.add_route("index", "/")
    config.add_route(
        "openapi_spec",
        "/openapi.json",
        factory="channelstream.wsgi_views.wsgi_security:AdminAuthFactory",
    )
    config.add_route(
        "admin",
        "/admin",
        factory="channelstream.wsgi_views.wsgi_security:AdminAuthFactory",
    )
    config.add_route(
        "admin_json",
        "/admin/admin.json",
        factory="channelstream.wsgi_views.wsgi_security:AdminAuthFactory",
    )
    config.add_route(
        "admin_action",
        "/admin/{action}",
        factory="channelstream.wsgi_views.wsgi_security:AdminAuthFactory",
    )
    # Backend API
    config.add_route("connect", "/connect")
    config.add_route("subscribe", "/subscribe")
    config.add_route("unsubscribe", "/unsubscribe")
    config.add_route("user_state", "/user_state")
    config.add_route("message", "/message")
    config.add_route("channel_config", "/channel_config")
    config.add_route("info", "/info")

    # listening API
    config.add_route("api_listen", "/listen")
    config.add_route("api_listen_ws", "/ws")
    config.add_route("api_disconnect", "/disconnect")

    config.add_route("section_action", "/{section}/{action}")


def handle_CORS(request):
    return ""



from gevent import monkey

monkey.patch_all()

import argparse
import copy
import logging
import pprint
import os
import configparser

import gevent
from gevent.server import StreamServer

import channelstream.wsgi_app as pyramid_app
import channelstream
from channelstream.cli import CONFIGURABLE_PARAMS, SHARED_DEFAULTS
from channelstream.gc import gc_conns_forever, gc_users_forever
from channelstream.policy_server import client_handle
from channelstream.ws_app import ChatApplicationSocket
from channelstream.utils import set_config_types

from ws4py.server.geventserver import WSGIServer
from ws4py.server.wsgiutils import WebSocketWSGIApplication

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class RoutingApplication(object):
    def __init__(self, config):
        self.ws_app = WebSocketWSGIApplication(handler_cls=ChatApplicationSocket)
        self.wsgi_app = pyramid_app.make_app(config)

    def __call__(self, environ, start_response):
        if environ["PATH_INFO"] == "/ws":
            environ["ws4py.app"] = self
            return self.ws_app(environ, start_response)

        return self.wsgi_app(environ, start_response)


def main():
    config = copy.deepcopy(SHARED_DEFAULTS)

    parser = argparse.ArgumentParser(add_help=True)
    parser.add_argument(
        "-i", "--ini", dest="ini", help="Config file path", default=None
    )
    parser.add_argument(
        "-v",
        "--version",
        dest="version",
        help="Print version info and exit",
        default=None,
        nargs="*",
    )
    args = parser.parse_args()

    if args.version is not None:
        print(channelstream.__version__)
        exit()

    ini_path = args.ini or os.environ.get("CHANNELSTREAM_INI")

    # set values from ini/cli
    if ini_path:
        parser = configparser.ConfigParser()
        parser.read(ini_path)
        settings = dict(parser.items("channelstream"))
        for key in CONFIGURABLE_PARAMS:
            try:
                config[key] = settings[key]
            except KeyError:
                pass
    else:
        for key in CONFIGURABLE_PARAMS:
            conf_value = os.environ.get(f"channelstream_{key}".upper())
            if conf_value is not None:
                config[key] = conf_value

    config = set_config_types(config)
    log_level = getattr(logging, (config.get("log_level") or "INFO").upper())
    logging.basicConfig(level=log_level)
    log.setLevel(log_level)
    log.debug(pprint.pformat(config))
    log.info("Starting channelstream {}".format(channelstream.__version__))
    url = "http://{}:{}".format(config["host"], config["port"])

    log.info("Starting flash policy server on port 10843")
    gevent.spawn(gc_conns_forever)
    gevent.spawn(gc_users_forever)
    server = StreamServer(("0.0.0.0", 10843), client_handle)
    server.start()
    log.info("Serving on {}".format(url))
    log.info("Admin interface available on {}/admin".format(url))
    if config["secret"] == "secret":
        log.warning("Using default secret! Remember to set that for production.")
    if config["admin_secret"] == "admin_secret":
        log.warning("Using default admin secret! Remember to set that for production.")
    log.warning(f"IP's allowed to post to API {config['allow_posting_from']}")
    server = WSGIServer(
        (config["host"], config["port"]),
        RoutingApplication(config),
        log=logging.getLogger("channelstream.WSGIServer"),
    )
    server.serve_forever()
